#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
#Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.

#This program is free software; you can redistribute it and/or modify
#it under the terms of the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.

#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#Apache License, Version 2.0 for more details.

import os
import io
import sys
import logging
import hpproto
from operator import attrgetter
from merge_proto_module import *
import copy


java_keyword_names = ["BUILDPARTIAL", "CLEAR", "CLEARFIELD", "CLEARONEOF", "CLONE", "EQUALS", "FINDVALUEBYNUMBER",
                    "FORNUMBER", "GETDEFAULTINSTANCE", "GETDEFAULTINSTANCEFORTYPE", "GETDESCRIPTOR",
                    "GETDESCRIPTORFORTYPE", "GETNUMBER", "GETPARSERFORTYPE", "GETSERIALIZEDSIZE", "GETUNKNOWNFIELDS",
                    "GETVALUEDESCRIPTOR", "HASHCODE", "INTERNALGETFIELDACCESSORTABLE", "INTERNALGETVALUEMAP",
                    "INTERNALVALUEMAP", "ISINITIALIZED", "MAYBEFORCEBUILDERINITIALIZATION", "MERGEFROM",
                    "MERGEUNKNOWNFIELDS", "NEWBUILDER", "NEWBUILDERFORTYPE", "PARSEDELIMITEDFROM", "PARSEFROM",
                    "PARSEPARTIALFROM", "PARSER", "REGISTERALLEXTENSIONS", "SETFIELD", "SETREPEATEDFIELD",
                    "SETUNKNOWNFIELDS", "TOBUILDER", "VALUEOF", "WRITETO", "DEFAULTINSTANCE", "DEFAULTINSTANCEFORTYPE",
                      "DESCRIPTOR", "DESCRIPTORFORTYPE", "NUMBER",
                     "PARSERFORTYPE", "SERIALIZEDSIZE", "UNKNOWNFIELDS", "VALUEDESCRIPTOR", "FIELD", "REPEATEDFIELD",
                     "UNKNOWNFIELDS"]


class Proto_writer():
    def __init__(self, proto_file):
        self.proto_file = proto_file
        self.crt_line_num = 1
        self.last_annotation_num = 0

    def sort_services(self, services):
        if type(services) == list and services != []:
            services.sort(key=attrgetter('name'))
        # sort statements in services
        for service in services:
            unsort_statement = []
            sorted_statement = []
            if type(service.statement) == list and service.statement != []:
                for statement in service.statement:
                    if statement.num == '':
                        unsort_statement.append(statement)
                    else:
                        sorted_statement.append(statement)
                if unsort_statement:
                    unsort_statement.sort(key=attrgetter('name'))
                    for tmp in unsort_statement:
                        service.statement.remove(tmp)
                        service.statement.append(tmp)
                crt_num = -1
                for statement in service.statement:
                    crt_num = self.get_current_num(crt_num, statement)
                    if statement.num == '':
                        statement.num = str(crt_num)
                    crt_num += 1
        return True

    def service_statement_str(self, scope, type, name, request, response):
        str = ''
        if scope is not None and scope != '' and scope != 'simple':
            str = ''.join([str, scope, ' '])
        if type is not None and type!='':
            str = ''.join([str, type, ' '])
        if name is not None and name!='':
            str = ''.join([str, 'rpc ', name, ' '])
        if request is not None and request!='':
            str = ''.join([str, '(', request, ') '])
        if response is not None and response!='':
            str = ''.join([str, 'returns (', response, ')'])
        str = ''.join([str, ';\n'])
        return str

    # output services
    def output_services(self, file, services):
        self.sort_services(self.proto_file.i_services)
        for service in services:
            # output structure of service
            if service.is_print == False:
                str = self.message_struct_str(service.keyword, service.name)
                self.file_write(file, str)
                service.is_print = True
            # output statement of service
            space = '   '
            # output statement of service
            for statement in service.statement:
                str = self.service_statement_str(statement.scope, statement.type,
                                                 statement.name, statement.Request, statement.Response)
                self.file_write(file, space+str)
            self.file_write(file, '}\n')
            # output requests of service
            request_definions = {}
            response_definions = {}
            for statement in service.statement:
                request_definions[statement.Request] = self.proto_file.definions[statement.Request]
                response_definions[statement.Response] = self.proto_file.definions[statement.Response]
            self.output_definions(request_definions, file, '')
            # output responses of service
            self.output_definions(response_definions, file, '')
        return True

    def file_write(self, file, str):
        file.write(str)
        self.crt_line_num += 1
        return True

    def output_annotations_begin(self, file, annotations):
        i = 0
        for i in range(self.last_annotation_num, len(annotations)):
            if annotations[i].pos.line > self.crt_line_num:
                break
            else:
                if annotations[i].pos.uses_pos is None:
                    self.file_write(file, annotations[i].str)
                    annotations[i].pos.uses_pos = self.crt_line_num
        self.last_annotation_num = i
        self.file_write(file, '\n')
        return True

    def output_annotations_end(self, file, annotations):
        i = 0
        for i in range(self.last_annotation_num, len(annotations)):
            if annotations[i].pos.uses_pos is None:
                self.file_write(file,annotations[i].str)
                annotations[i].pos.uses_pos = self.crt_line_num
        self.last_annotation_num = i
        self.file_write(file, '\n')
        return True

    def get_public_import(self, import_str):
        imports = import_str.split(' ')
        if len(imports) == 2 and imports[0] == 'public':
            return imports[1]
        else:
            return ''

    # output heads
    def output_heads(self, file):
        syntax = 'proto3'
        # print syntex
        if self.proto_file.i_syntax is not None and self.proto_file.i_syntax != '':
            syntax = self.proto_file.i_syntax
        self.file_write(file, 'syntax = "%s";\n' % syntax)
        # print annotations
        self.output_annotations_begin(file, self.proto_file.i_annotations)
        # print imports
        for mpt in self.proto_file.i_imports:
            public_import = self.get_public_import(mpt)
            if public_import != '':
                self.file_write(file, 'import public "%s";\n' % public_import)
            else:
                self.file_write(file, 'import "%s";\n' % mpt)
        # print package
        if self.proto_file.i_package is not None:
            self.file_write(file, 'package %s;\n' % self.proto_file.i_package)
            self.file_write(file, '\n')
        # print service
        self.output_services(file, self.proto_file.i_services)
        return True

    def get_current_num(self, num, sub_msg):

        # arrange oneof num
        if sub_msg.num != '':
            crt_num = int(sub_msg.num)
            return crt_num
        crt_num = num + 1
        if num == -1:  # set the initial value for num
            if sub_msg.parent is not None and sub_msg.parent.keyword == 'enum':
                crt_num = 0  # initial value for enum element
            else:
                crt_num = 1  # initial value for others
        return crt_num

    def insert_dummy_enum_in_first(self, statements):
        enum_prefix = statements[0].name.split('_')[0]
        dummy_enum = copy.copy(statements[0])
        dummy_enum.enum_val = 0
        dummy_enum.name = '%s_INVALID_ENUM_VALUE' % enum_prefix
        statements.insert(0, dummy_enum)



    def sort_enum_statement(self, statements):
        #sort the enum by value
        statements.sort(key=attrgetter('enum_val'))
        if statements[0].enum_val != 0:
            self.insert_dummy_enum_in_first(statements)

    def assign_old_proto_num_to_yang_order_num(self, old_proto_statement):
        for old_stat in old_proto_statement:
            old_stat.yang_node_order_num = int(old_stat.num)

    def sort_statement_by_key(self, statement, parent_statements):
        statement.sort(key=attrgetter('yang_node_order_num'))
        for stat in statement:
            parent_statements.remove(stat)
            parent_statements.append(stat)

    # sort statements
    def sort_statements(self, statements, crt_num=-1):
        unsort_statement = []
        old_proto_statement = []
        # current max num
        max_num = crt_num
        if type(statements) == list and statements != []:
            for statement in statements:
                if statement.num == '':
                    unsort_statement.append(statement)
                else:
                    if statement.num and int(statement.num) > max_num:
                        max_num = int(statement.num)
                    old_proto_statement.append(statement)
            #the node with smaller proto num maybe occurs behind the node with greater proto num, so need resort the nodes
            #according proto num of old proto
            self.assign_old_proto_num_to_yang_order_num(old_proto_statement)
            self.sort_statement_by_key(old_proto_statement, statements)

            # if the proto num of nodes are empty, sort them according the sequence of yang file
            self.sort_statement_by_key(unsort_statement, statements)

            # 对enum排序：1.有old proto时，先按old proto顺序排序；2.没有old proto时，按enum value排序
            if statements[0].parent.keyword == 'enum':
                self.sort_enum_statement(statements)
                #self.assign_enum_value_to_num(statements)
            statement_len = len(statements)
            for i in range(0, statement_len):
                if type(statements[i].type) == hpproto.proto_proto.DefinionObj \
                             and statements[i].type.keyword == 'oneof':
                    continue
                crt_num = self.get_current_num(crt_num, statements[i])
                if not statements[i].num:
                    crt_num = max(crt_num, max_num+1)
                    max_num = crt_num
                    statements[i].num = str(crt_num)
                else:
                    # enum or oneof ignore
                    if statements[i].parent.keyword == 'enum':
                        continue
                    # 多个proto文件合并时，可能出现后面statement的num值小于前面的情形，此时给后面num值+1处理
                    if i+1 < statement_len and statements[i+1].num != '':
                        last_stat_num = int(statements[i].num)
                        next_stat_num = int(statements[i+1].num)
                        if next_stat_num <= last_stat_num:
                            next_stat_num = last_stat_num + 1
                            statements[i+1].num = str(next_stat_num)
                        crt_num = last_stat_num
                if type(statements[i].type) == hpproto.proto_proto.DefinionObj :
                    if statements[i].type.keyword == 'oneof':
                        self.sort_statements(statements[i].type.statement, max_num)
                        if statements[i].type.statement:
                            max_num = max(int(statements[i].type.statement[-1].num), max_num)
                            crt_num = max_num
                    else:
                        self.sort_statements(statements[i].type.statement)
        return True

    # sort definions
    def sort_definions(self, definions):
        if not definions:
            return True
        for definion in definions.values():
            if definion.statement:
                self.sort_statements(definion.statement)
        return True

    def message_struct_str(self, keyword, str_type):
        if str_type == 'Empty':
            str_out = '%s Empty { ' % keyword
        else:
            str_out = '%s %s {\n' % (keyword, str_type)
        return str_out

    def message_statement_str(self, statement):
        str_out = ''
        yang_node_name = statement.yang_node_name
        scope = statement.scope
        if isinstance(statement.type, hpproto.proto_proto.DefinionObj):
            type_str = statement.type.name
        else:
            type_str = statement.type
        name = statement.name
        num = statement.num
        proto_annotations = statement.proto_annotations
        if scope and scope != 'simple':
            str_out = '%s%s ' % (str_out, scope)
        if type_str :
            str_out = '%s%s ' % (str_out, type_str)
        if name :
            str_out = '%s%s ' % (str_out, name)
        if yang_node_name != '':
            if yang_node_name.startswith('[') and yang_node_name.endswith(']'):  # old proto中有[]，直接将[]弄过来
                yang_node_str = ' %s' % yang_node_name
            else:
                yang_node_str = ' [json_name = "%s"]' % yang_node_name

            if proto_annotations:
                str_out = '%s=%s%s; ' % (str_out, num, yang_node_str)
                for annotation in proto_annotations:
                    str_out = '%s//%s' % (str_out, annotation)
            else:
                str_out = '%s= %s%s;\n' %(str_out, num, yang_node_str)
        else:
            str_out = '%s= %s;\n' % (str_out, num)
            if statement.parent.keyword != 'enum':
                logging.info('node %s json name is null, xpath:%s' % (statement.name, get_proto_xpaths(statement)))
        return str_out

    # output statement
    def output_statement(self, statement, file, definions, space=''):
        if not statement:
            return True
        for sub_statement in statement:
            if type(sub_statement.type) != hpproto.proto_proto.DefinionObj:
                if sub_statement.name != '':
                    str_out = self.message_statement_str(sub_statement)
                    self.file_write(file, ''.join([space, str_out]))
            elif sub_statement.type.keyword == 'oneof':
                # out_put sub_message out of oneof
                if sub_statement.type.statement:
                    self.output_statement(sub_statement.type.statement, file, sub_statement.type.definions,
                                          space)
                    str_out = self.message_struct_str(sub_statement.type.keyword, sub_statement.type.name)
                    self.file_write(file, ''.join([space, str_out]))
                    for sub_stmt in sub_statement.type.statement:
                        str_line = self.message_statement_str(sub_stmt)
                        if str_line:
                            self.file_write(file, ''.join(['%s   ' % space, str_line]))
                    self.file_write(file, '%s}\n' % space)
                    sub_statement.type.is_print = True
            else:
                tail_str = ''
                # output structure of type
                if sub_statement.type.keyword in hpproto.proto_proto.KEYWORD and not sub_statement.type.is_print:
                    str_out = self.message_struct_str(sub_statement.type.keyword, sub_statement.type.name)
                    self.file_write(file, ''.join([space, str_out]))
                    sub_statement.type.is_print = True
                    if sub_statement.type.statement:
                        sub_space = '%s   ' % space
                        self.output_statement(sub_statement.type.statement, file, sub_statement.type.definions,
                                              sub_space)
                    tail_str = '%s}' % space
                    if sub_statement.type.keyword == 'enum':
                        tail_str = '%s;' % tail_str
                # output tail of structure
                if tail_str != '':
                    self.file_write(file, '%s\n'% tail_str)
                # output statement
                # case do not need output outside
                if sub_statement.name != '' and sub_statement.parent.keyword != 'oneof':
                    str_line = self.message_statement_str(sub_statement)
                    if str_line:
                        self.file_write(file, ''.join([space, str_line]))

        self.output_definions(definions, file, space)
        return True

    # output definions
    def output_definions(self, definions, file, space=''):
        if not definions:
            return True
        sorted_definions = list(definions.values())
        sorted_definions.sort(key=attrgetter('pos.line'))
        unsort_definions = []
        if type(sorted_definions) == list and sorted_definions != []:
            for tmp in sorted_definions:
                if tmp.pos.line == 0:
                    unsort_definions.append(tmp)
            if unsort_definions:
                unsort_definions.sort(key=attrgetter('yang_node_order_num'))

                for tmp in unsort_definions:
                    sorted_definions.remove(tmp)
                    sorted_definions.append(tmp)
        # output the root definions fist
        if self.proto_file.i_root_elements:
            root_definions = []
            for i in range(len(sorted_definions)-1, -1, -1):
                tmp = sorted_definions[i]
                if tmp.name in self.proto_file.i_root_elements:
                    root_definions.insert(0, tmp)
                    sorted_definions.pop(i)

            if root_definions:
                sorted_definions = root_definions + sorted_definions
        for definion in sorted_definions:
            if definion.is_print:
                continue
            if definion.is_top:
                space = ''
            tail_str = ''
            # output definions
            if definion.keyword in hpproto.proto_proto.KEYWORD:
                str = self.message_struct_str(definion.keyword, definion.name)
                self.file_write(file, ''.join([space, str]))
                definion.is_print = True
                # output sub statement
                if definion.statement:
                    sub_space = '%s   ' % space
                    self.output_statement(definion.statement, file, definion.definions, sub_space)
                if definion.name == 'Empty':
                    tail_str = '}'
                else:
                    tail_str = '%s}' % space
                if definion.keyword == 'enum':
                    tail_str = '%s;' % tail_str
            # output tail of definions
            if tail_str != '':
                self.file_write(file, '%s\n' % tail_str)

        return True

    # edit profile for test
    def edit_profile(self, proto_file):
        self.proto_file.definions['XQoS4QueueResPreAlarmCancelTrap'].pos.line = 0
        self.proto_file.definions['HqosSQApplyPirFailAlarmTrap'].pos.line = 0
        return True

    def first_letter_upper(self, name):
        # 当definion为'_24GHz'时，split('_')后传入的name为空，会产生异常
        strs = ''
        try:
            strs = ''.join([name[0].upper(), name[1:]])
        except:
            pass

        return strs

    def first_attr_upper(self, name):
        if name[0].isupper():
            return True
        return False

    def find_and_delete_invalid_definions(self, statements, definions):
        type_names = []
        statements_names = [stat.name for stat in statements]
        # delete old_proto same_name_node
        del_stat_dict = {}
        cap_statements_names = [stat_name.replace('_', '').upper() for stat_name in statements_names]
        from collections import defaultdict
        del_stats = []
        for java_keyword_name in java_keyword_names:
            if java_keyword_name in cap_statements_names:
                del_stats.append(statements[cap_statements_names.index(java_keyword_name)])
                logging.info('remove node %s, because it is java key word'
                             % statements[cap_statements_names.index(java_keyword_name)].name)
        d = defaultdict(list)
        for k, va in [(v, i) for i, v in enumerate(cap_statements_names)]:
            d[k].append(va)
        # oldproto为人工书写的，遇到特殊字符转成了驼峰格式，与目前工具转换逻辑不一致，所以取工具转换节点名
        # 目前通过下划线个数来判断，个数多的就是工具转换的，目前仅支持同一个父节点下出现2个，认为匹配成功
        for values in d.values():
            if len(values) == 2:
                if statements_names[values[0]].count('_') > statements_names[values[1]].count('_'):
                    del_stat_dict[values[0]] = values[1]
                else:
                    del_stat_dict[values[1]] = values[0]

        for new_index in del_stat_dict.keys():
            old_stat = statements[del_stat_dict[new_index]]
            new_stat = statements[new_index]
            new_stat.num = old_stat.num
            del_stats.append(old_stat)
            logging.info("remove node %s, because it is same with node %s", str(old_stat.name), str(new_stat.name))
        del_stats = set(del_stats)
        for stat in del_stats:
            statements.remove(stat)

        for stat in statements:
            if isinstance(stat.type, hpproto.proto_proto.DefinionObj):
                if stat.type.statement:
                    self.find_and_delete_invalid_definions(stat.type.statement, stat.type.definions)

                if stat.yang_node_name != '' or stat.type.pos.line != 0:
                    type_names.append(stat.type.name)

        if len(definions) > 0:
            del_keys = []
            for definion_name in definions.keys():
                # deal with empty oneof
                if isinstance(definions[definion_name], hpproto.proto_proto.DefinionObj):
                    if definions[definion_name].keyword == 'oneof' and not definions[definion_name].statement:
                        del_keys.append(definion_name)
                        continue
                if definion_name not in type_names and definion_name not in del_keys:
                    del_keys.append(definion_name)
                elif definion_name.replace('_', '').upper() in java_keyword_names:
                    del_keys.append(definion_name)

            for del_key in del_keys:
                logging.info('delete invalid definion:%s' % del_key)
                definions.pop(del_key)

    def delete_invalid_definions(self, definions):
        for definion in definions.values():
            self.find_and_delete_invalid_definions(definion.statement, definion.definions)

    def output_proto_file(self, filename=None):
        if not filename:
            #filename = os.path.join(self.out_directory, '%s.proto' % self.module_name)
            filename = self.proto_file.ref
        logging.info(
            "==============================START OUPUT PROTO %s===================================" % filename)
        with open(filename, "w", encoding="utf-8") as file:
            self.output_heads(file)
            self.delete_invalid_definions(self.proto_file.definions)
            self.sort_definions(self.proto_file.definions)
            if self.proto_file.definions:
                for proto_file_definion in self.proto_file.definions.values():
                    del_same_node_in_proto(proto_file_definion)
            self.output_definions(self.proto_file.definions, file)
            self.output_annotations_end(file, self.proto_file.i_annotations)
        return True


#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
#Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.

#This program is free software; you can redistribute it and/or modify
#it under the terms of the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.

#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#Apache License, Version 2.0 for more details.


import re
import sys
import hpproto
import time
import shutil
import proto_writer
from optparse import OptionParser
import pyang
from pyang.error import Position
from hpproto.proto_proto import MessageObj
from hpproto.proto_proto import AnnotationObj
from hpproto.proto_proto import ProtoFile
import yang_util
from merge_proto_module import *

rep_revs_log_file = 'convert_yang2proto'
yang_root_node_keyword = 'container'
yang_top_node_keywords = ['notification', 'anydata', 'anyxml', 'container']
enum_type = ['bit', 'enumeration']
message_type = ['anydata', 'anyxml', 'augment', 'case', 'container', 'notification']
repeated_type = ['leaf-list', 'list']
oneof_type = ['choice']
yang_keyword_not_support = ['action']
proto_version = 'proto3'
yangtype_to_prototype = {
    'binary': 'bytes',
    'bits': 'enum',
    'boolean': 'bool',
    'empty': 'bool',
    'enumeration': 'enum',
    'instance-identifier': 'string',
    'int8': 'int32',
    'int16': 'int32',
    'int32': 'int32',
    'int64': 'int64',
    'string': 'string',
    'uint8': 'uint32',
    'uint16': 'uint32',
    'uint32': 'uint32',
    'uint64': 'uint64',
    'union': 'string',
    'identityref': 'string',
    'decimal64': 'Decimal64'
}
enum_val_format = ["!", '"', "#", "￥", "$", "%", "&", ",", "-", "(", ")", "*", "+", "/", ":", "<", ">", "=", ".",
                   "?", "[", "]", "{", "}", "|", "^", "~", "`", "@", " ", ";"]
name_not_support = []
same_name_suffix = '_1'
number = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '_']
origin_keep_word = ["BUILDPARTIAL", "CLEAR", "CLEARFIELD", "CLEARONEOF", "CLONE", "EQUALS", "FINDVALUEBYNUMBER",
                    "FORNUMBER", "GETDEFAULTINSTANCE", "GETDEFAULTINSTANCEFORTYPE", "GETDESCRIPTOR",
                    "GETDESCRIPTORFORTYPE", "GETNUMBER", "GETPARSERFORTYPE", "GETSERIALIZEDSIZE", "GETUNKNOWNFIELDS",
                    "GETVALUEDESCRIPTOR", "HASHCODE", "INTERNALGETFIELDACCESSORTABLE", "INTERNALGETVALUEMAP",
                    "INTERNALVALUEMAP", "ISINITIALIZED", "MAYBEFORCEBUILDERINITIALIZATION", "MERGEFROM",
                    "MERGEUNKNOWNFIELDS", "NEWBUILDER", "NEWBUILDERFORTYPE", "PARSEDELIMITEDFROM", "PARSEFROM",
                    "PARSEPARTIALFROM", "PARSER", "REGISTERALLEXTENSIONS", "SETFIELD", "SETREPEATEDFIELD",
                    "SETUNKNOWNFIELDS", "TOBUILDER", "VALUEOF", "WRITETO"]
ext_keep_word = ["DEFAULTINSTANCE", "DEFAULTINSTANCEFORTYPE", "DESCRIPTOR", "DESCRIPTORFORTYPE", "NUMBER",
                     "PARSERFORTYPE", "SERIALIZEDSIZE", "UNKNOWNFIELDS", "VALUEDESCRIPTOR", "FIELD", "REPEATEDFIELD",
                     "UNKNOWNFIELDS"]
# space not allowed in yang rfc
compile_str = '// compiled %s\n' % time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(time.time()))
annotations = [
    '// Automatically generated by ProtoYang\n',
    compile_str, '// ProtoYang-version 1\n',
]


def env_parse():
    parser = OptionParser()
    parser.add_option("-Y", "--yang", dest="yang_path", default=r'./yang',
                      help="The directory of yang files or a single yang file.",
                      metavar='"COMMON YANG FILE"')
    parser.add_option("-O", "--out", dest="out_proto_directory", default=r'./proto',
                      help="The directory for generated .proto files",
                      metavar='"ADD FILE DIRECTORY"')
    parser.add_option("-S", "--src", dest="src_proto_directory", default=r'./src_proto',
                      help="The source .proto file directory",
                      metavar='"SOURCE PROTO FILE DIRECTORY"')
    parser.add_option("-L", "--logdir", dest="log_directory", default=r'./logs',
                      help="The direcotry to create the log file",
                      metavar='"LOG DIRECTORY"')
    parser.add_option("-D", "--dependencies", dest="yang_dependencies", default=r'./yang_dependencies',
                      help="The directory used as yang dependencies ",
                      metavar='"DEPENDENCIES DIR"')
    (options, args) = parser.parse_args()
    # for directory in options.yang_path, \
    #                  options.out_proto_directory, \
    #                  options.log_directory:
    #     if not directory or not os.path.isdir(directory):
    #         sys.stderr.write("[0x000001] :directory doesn't exists: %s\n" % directory)
    #         sys.exit(1)
    if not os.path.exists(options.yang_path):
        sys.stderr.write("[0x000001] :yang path doesn't exists: %s\n" % options.yang_path)
        sys.exit(1)
    return options

class Yang_proto_converter():
    def __init__(self, options):
        self.yang_path = options.yang_path
        self.out_proto_directory = options.out_proto_directory
        self.log_directory = options.log_directory
        self.src_proto_directory = options.src_proto_directory
        self.yang_dependencies = options.yang_dependencies
    def translate_services(self):
        return

    def split_same_name_inonestruct(self, structs):
        names = []
        stat_list = structs.statement
        for stat in stat_list[:]:
            if stat.type is not None:
                if type(stat.type) is not str:
                    if stat.name == stat.type.name:
                        stat.type.name = ''.join([stat.name, same_name_suffix])
                    self.split_same_name_inonestruct(stat.type)
                else:
                    if stat.name == stat.type:
                        stat.name = ''.join([stat.name, same_name_suffix])
            if stat.name in names:
                # 当enum中出现重复时，直接删除重复的enum，不进行重命名
                # 遇到有重复的enum，按yang文件顺序从前往后删除重复的enum
                if structs.keyword == 'enum':
                    logging.info('delete duplicate enum:%s, xpath:%s' % (stat.name, get_proto_xpaths(stat)))
                    stat_list.remove(stat)
                    continue
                else:
                    stat.name = ''.join([stat.name, same_name_suffix])

            names.append(stat.name)

    def add_decimal64_definion(self, proto_file_name, proto_node, proto_file):
        if self.check_proto_has_decimal64(proto_node):
            self.translate_decimal64_to_proto(proto_file_name, proto_file)

    def check_proto_has_decimal64(self, proto_node):
        if not proto_node:
            return False
        for def_obj in proto_node.values():
            if self.check_proto_has_decimal64(def_obj.definions):
                return True
            for stat in def_obj.statement:
                if stat.type == 'Decimal64':
                    return True
                elif isinstance(stat.type, DefinionObj):
                    if self.check_proto_has_decimal64(stat.type.definions):
                        return True
        return False

    def _do_find_augment_node_parent(self, nodes, pos, stat):
        if pos == 0:
            return stat

        if stat.parent.arg == nodes[pos - 1]:
            stat = stat.parent
            pos -= 1
            return self._do_find_augment_node_parent(nodes, pos, stat)

    def find_augment_node_parent(self, xpath, nodes, yang_node):
        for stat in yang_node.substmts:
            if stat.arg == xpath and stat.i_target_node is not None:
                return self._do_find_augment_node_parent(nodes, len(nodes) - 1, stat.i_target_node)

    def get_module_full_name(self, module_prefix, yang_node):
        return yang_node.i_prefixes[module_prefix][0]

    def translate_augment(self, yang_node, ptx):
        use_augment = False
        augment_path = []
        module_prefixs = []
        nodes_str = []
        for stat in yang_node.substmts:
            if stat.keyword == 'augment':
                augment_path.append(stat.arg)
                use_augment = True

        if use_augment is not True:
            return

        augmented_module_prefix_list = []
        for xpath in augment_path:
            module_prefixs.clear()
            nodes_str.clear()

            logging.info('augment: current module:%s, xpath:%s', yang_node.i_modulename, xpath)
            pattern_module = re.compile(r'/[\w\-]*:[\w\-]*')
            m = pattern_module.findall(xpath)
            if m:
                for match_path in m:
                    module_prefix_str = match_path.split(':')[0].strip('/')
                    node_str = match_path.split(':')[1]
                    module_prefixs.append(module_prefix_str)
                    nodes_str.append(node_str)
            else:
                logging.info('xpath no match:%s', xpath)
                continue

            augmented_module_prefix = module_prefixs[0]
            if augmented_module_prefix not in augmented_module_prefix_list:
                augmented_module_prefix_list.append(augmented_module_prefix)
            else:
                continue

            parent_node = self.find_augment_node_parent(xpath, nodes_str, yang_node)
            if parent_node is None:
                logging.error('xpath(%s) not found parent node.' % xpath)
                continue

            # 将augment的模块写成aug临时文件
            # 可能存在augment多个原始yang文件情形，比如a.yang同时对b.yang和c.yang进行augment
            # augment临时文件命名：augment module name+ '-' + '原始modulename' + '_aug.proto'
            top_parent_node = parent_node.parent
            if top_parent_node is not None:
                proto_file_name = os.path.join(self.out_directory, '%s-%s_aug.proto'
                                               % (yang_node.i_modulename,top_parent_node.i_modulename))
                proto_file = ProtoFile(proto_file_name, proto_version)
                self.translate_heads(proto_file_name, top_parent_node, proto_file)
                self.translate_services()
                self.translate_messages(proto_file_name, top_parent_node, proto_file)
                old_proto_file_name = os.path.join(self.src_proto_directory, '%s.proto'% top_parent_node.i_modulename)
                m = self.get_module_for_out(old_proto_file_name, ptx, top_parent_node.i_modulename, proto_file)
                if ptx.errors:
                    return
                if m:
                    m.output_proto_file(proto_file_name)
                logging.info('generate tmp augment proto file(%s)', proto_file_name)
        return

    def find_top_yang_node_order_num(self, node, top_nodes):
        i = 0
        top_nodes_len = len(top_nodes)
        for i in range(0, top_nodes_len):
            if top_nodes[i].arg == node.arg:
                break

        return i

    def translate_messages(self, proto_file_name, yang_node, proto_file):
        top_nodes = yang_node.i_children
        for node in top_nodes:
            if node.keyword not in yang_top_node_keywords:
                continue
            proto_module = DefinionObj(proto_file_name)

            proto_module.keyword = 'message'
            name = self.get_out_message_definion_name(node.arg, '')
            proto_module.name = name
            proto_module.proto_xpath = '/%s' % name
            proto_module = self._parse_statement(proto_file_name, proto_module, node)

            self.split_same_name_inonestruct(proto_module)
            proto_file.definions[name] = proto_module
            if node.keyword == yang_root_node_keyword:
                proto_file.i_root_elements.append(name)

            proto_module.yang_node_order_num = self.find_top_yang_node_order_num(node, top_nodes)
        self.add_decimal64_definion(proto_file_name, proto_file.definions, proto_file)

    def get_state_from_module(self, yang_module):
        if yang_module.keyword in message_type:
            return 'message'
        if yang_module.keyword in repeated_type:
            return'repeated'
        if yang_module.keyword in oneof_type:
            return 'oneof'
        else:
            return 'simple'

    def first_letter_upper(self, name):
        # 当definion为'_24GHz'时，split('_')后传入的name为空，会产生异常
        strs = ''
        try:
            strs = ''.join([name[0].upper(), name[1:]])
        except:
            pass

        return strs

    def first_attr_upper(self, name):
        if name[0].isupper():
            return True
        return False

    def get_out_message_definion_name(self, name, father_proto_xpath):
        if '_' in name:
            for attr in name.split('_'):
                if not attr.istitle():
                    name = name.replace('_%s' % attr, self.first_letter_upper(attr))
        if '-' in name:
            for attr in name.split('-'):
                if not self.first_attr_upper(attr):
                    name = name.replace('-%s' % attr, self.first_letter_upper(attr))
                else:
                    name = name.replace('-%s' % attr, '_%s'% self.first_letter_upper(attr))
        index = 0
        for str in name:
            index += 1
            if str in number and index < len(name):
                name = name.replace(''.join([str, name[index]]), ''.join([str, name[index].upper()]))
        name = self.first_letter_upper(name)
        # calculate if same node occured
        name = self.calculate_proto_name(father_proto_xpath, name)
        return name

    def get_enumval_type(self, node_type, enum_val, father_proto_xpath):
        for enum_spec in enum_val_format:
            if enum_spec in enum_val:
                enum_val = enum_val.replace(enum_spec, 'X_')

        for enum_spec in number:
            if enum_val.startswith(enum_spec):
                enum_val = 'X_%s' % enum_val
                break
        if '-' in enum_val:
            enum_val = enum_val.replace('-', '_')
        enum_name = self.calculate_proto_name(father_proto_xpath, '%s_%s' % (node_type, enum_val.upper()))
        return enum_name

    def get_type_name_for_leaf(self, yang_module):
        try:
            type_module = yang_module.search_one('type')
            if type_module.i_type_spec is None:
                logging.error('type %s in %s line %s can not be find', type_module.arg,
                              os.path.basename(type_module.pos.ref), type_module.pos.line)
                return ''
            if yang_module.i_leafref is None:
                return yangtype_to_prototype[type_module.i_type_spec.name]
            else:
                return self.get_type_name_for_leaf(type_module.i_type_spec.i_target_node)
        except Exception:
            logging.error('get_type_name_for_leaf %s in module %s' % (yang_module.arg, self.module_name))

    def split_yang_node(self, node):
        results_node = {}
        results_node['message'] = []
        results_node['repeated'] = []
        results_node['oneof'] = []
        results_node['simple'] = []
        try:
            for i_node in node.i_children:
                results_node[self.get_state_from_module(i_node)].append(i_node)
        except:
            pass
        return results_node

    def _get_enum_value_exec(self, enum_name, enum_statement):
        enum_val = -1
        for stat in enum_statement.substmts:
            if stat.arg == enum_name:
                for enum_body in stat.substmts:
                    if enum_body.keyword == 'value':
                        enum_val = int(enum_body.arg)
                        break
                break
        return enum_val

    # 获取yang文件中enum的value，没有value时，赋值-1
    # 在enum name存在重复时，删除后，可以依据enum value排序
    def get_enum_value(self, enum_name, type_module):
        enum_val = -1
        if type_module.i_typedef is not None:
            enum_statement = type_module.i_typedef.search_one('type')
            if enum_statement is not None and enum_statement.arg == 'enumeration':
                enum_val = self._get_enum_value_exec(enum_name, enum_statement)
        elif type_module.arg == 'enumeration':
            enum_val = self._get_enum_value_exec(enum_name, type_module)
        return enum_val

    def translate_enum_to_proto(self, proto_module, enum_name, proto_file_name, yang_module):
        type_module = yang_module.search_one('type')
        if yang_module.i_leafref is not None:
            return self.translate_enum_to_proto(proto_module, enum_name, proto_file_name,
                                                type_module.i_type_spec.i_target_node)
        if type_module.i_type_spec.name == 'bits':
            for enum_val in type_module.i_type_spec.bits:
                sub_proto_module = MessageObj(proto_file_name)
                sub_proto_module.scope = 'simple'
                sub_proto_module.name = self.get_enumval_type(str(enum_name), str(enum_val[0]), proto_module.proto_xpath)
                sub_proto_module.parent = proto_module
                sub_proto_module.proto_xpath = '%s/%s' % (proto_module.proto_xpath, sub_proto_module.name)
                proto_module.statement.append(sub_proto_module)
        elif type_module.i_type_spec.name == 'enumeration':
            for enum_val in type_module.i_type_spec.enums:
                sub_proto_module = MessageObj(proto_file_name)
                sub_proto_module.scope = 'simple'
                sub_proto_module.name = self.get_enumval_type(str(enum_name), str(enum_val[0]), proto_module.proto_xpath)
                #sub_proto_module.enum_val = self.get_enum_value(str(enum_val[0]), type_module)
                sub_proto_module.enum_val = enum_val[1]
                if sub_proto_module.enum_val != -1:
                    sub_proto_module.num = str(sub_proto_module.enum_val)
                sub_proto_module.parent = proto_module
                sub_proto_module.proto_xpath = '%s/%s' % (proto_module.proto_xpath, sub_proto_module.name)
                proto_module.statement.append(sub_proto_module)

        return proto_module

    def translate_decimal64_to_proto(self, proto_file_name, proto_file):
        proto_module = DefinionObj(proto_file_name)
        proto_module.keyword = 'message'
        name = 'Decimal64'
        proto_module.name = name
        proto_file.definions[name] = proto_module

        # 例：int64  value = 1;
        int64_sub_proto_module = MessageObj(proto_file_name)
        int64_sub_proto_module.scope = 'simple'
        int64_sub_proto_module.name = 'value'
        int64_sub_proto_module.type = 'int64'
        int64_sub_proto_module.num = '1'
        int64_sub_proto_module.parent = proto_module
        int64_sub_proto_module.yang_node_name = 'value'
        proto_module.statement.append(int64_sub_proto_module)
        # 例：uint32 fraction_digits = 2;
        uint32_sub_proto_module = MessageObj(proto_file_name)
        uint32_sub_proto_module.scope = 'simple'
        uint32_sub_proto_module.name = 'fraction_digits'
        uint32_sub_proto_module.type = 'uint32'
        uint32_sub_proto_module.num = '2'
        uint32_sub_proto_module.parent = proto_module
        uint32_sub_proto_module.yang_node_name = 'fraction-digits'
        proto_module.statement.append(uint32_sub_proto_module)

        return proto_module

    def get_out_union_type(self, type_module):
        if type_module.i_type_spec.name:
            pass

    def trans_yangleaf_to_proto(self, proto_module, yang_module):
        proto_module.type = self.get_type_name_for_leaf(yang_module)
        if self.bits_scope_change(yang_module):
            proto_module.scope = 'repeated'
        return proto_module

    def bits_scope_change(self, yang_module):
        type_module = yang_module.search_one('type')
        if yang_module.i_leafref is None:
            try:
                if type_module.i_type_spec.name == 'bits':
                    return True
            except:
                logging.error("import in yang file:%s related to the node type is missing." % yang_module.i_module.arg)
        return False

    def trans_node_name(self, arg, father_proto_xpath, is_leaf_node=False):
        # calculate if same node occured
        arg = arg.replace('-', '_')
        name = self.calculate_proto_name(father_proto_xpath, arg, is_leaf_node)

        return name

    def trans_module_type(self, arg):
        return self.first_letter_upper(arg)

    # 使用uses grouping对choice进行augment，并且augment的内容是简写的case(参考RFC7950 7.9.2 choice的case声明)
    # （即分支包含单个的”anydata”、“anyxml”、”choice”、“container”、“leaf”、“list”或者“leaf-list”声明，”case”声明可以忽略）
    # example：这里leaf c-aaa等同于case c-aaa{ leaf c-aaa }
    #      例：uses choice:choice-subnet {
    #        例：augment "test-choice/subnet" {
    #            例：leaf c-aaa{
    #                例：type string;
    #            }
    #         }
    #       }
    def is_augment_shorthand_case(self, yang_node):
        if yang_node.keyword == 'case' and len(yang_node.i_children) == 1:
            yang_node_pos = yang_node.pos
            child_node_pos = yang_node.i_children[0].pos
            if yang_node_pos.line == child_node_pos.line and yang_node_pos.ref == child_node_pos.ref:
                if yang_node.i_children[0].i_augment.keyword == 'augment':
                    logging.info('augment shorthand case, yang_node:%s' % yang_node.arg)
                    return True
        return False

    def find_yang_node_order_num(self, yang_node):
        children_list = yang_node.parent.i_children
        i = 0
        for children in children_list:
            if children.arg == yang_node.arg:
                return i
            if children.keyword == 'choice':
                i += len(children.search('case'))
                continue
            i += 1
        return 0

    def arrange_message_type(self, container_yang_module, proto_file_name, proto_module):
        sub_message_name = self.trans_node_name(container_yang_module.arg, proto_module.proto_xpath)
        sub_message_type = self.get_out_message_definion_name(container_yang_module.arg, proto_module.proto_xpath)
        sub_define_module = DefinionObj(proto_file_name)
        sub_define_module.keyword = 'message'
        sub_define_module.parent = proto_module

        sub_define_module.name = sub_message_type
        sub_define_module.proto_xpath = '%s/%s' % (proto_module.proto_xpath, sub_message_type)
        yang_node_order_num = self.find_yang_node_order_num(container_yang_module)
        sub_define_module.yang_node_order_num = yang_node_order_num + proto_module.yang_node_order_num

        proto_module.definions[sub_message_type] = self._parse_statement(
            proto_file_name, sub_define_module, container_yang_module)

        sub_proto_module = MessageObj(proto_file_name)
        sub_proto_module.name = sub_message_name
        sub_proto_module.parent = proto_module
        sub_proto_module.proto_xpath = '%s/%s' % (proto_module.proto_xpath, sub_message_type)
        sub_proto_module.yang_node_order_num = yang_node_order_num
        sub_proto_module.yang_node_name = container_yang_module.arg

        sub_proto_module.type = proto_module.definions[sub_message_type]
        if container_yang_module.keyword == 'case' and proto_module.keyword == 'oneof':
            sub_define_module.name = '%s_%s' % (proto_module.name, sub_message_type)
            sub_define_module.proto_xpath = '%s/%s' % (proto_module.proto_xpath, sub_define_module.name)
            sub_proto_module.name = '%s_%s' %(container_yang_module.parent.arg.replace('-', '_'), sub_proto_module.name)
            sub_proto_module.proto_xpath = '%s/%s' % (proto_module.proto_xpath, sub_proto_module.name )
            sub_proto_module.yang_node_order_num = yang_node_order_num + proto_module.yang_node_order_num
            proto_module.parent.statement.append(sub_proto_module)
            proto_module.parent.definions[sub_define_module.name] = proto_module.definions[sub_message_type]
            sub_proto_module.type = proto_module.parent.definions[sub_define_module.name]
            del proto_module.definions[sub_message_type]

        proto_module.statement.append(sub_proto_module)
        return proto_module

    def arrange_repeated_type(self, container_yang_module, proto_file_name, proto_module):
        sub_message_name = self.trans_node_name(container_yang_module.arg, proto_module.proto_xpath)
        sub_message_type = self.get_out_message_definion_name(container_yang_module.arg, proto_module.proto_xpath)
        sub_proto_module = MessageObj(proto_file_name)
        sub_proto_module.scope = 'repeated'
        sub_proto_module.name = sub_message_name
        sub_proto_module.parent = proto_module
        sub_proto_module.proto_xpath = '%s/%s' % (proto_module.proto_xpath, sub_message_type)
        yang_node_order_num = self.find_yang_node_order_num(container_yang_module)
        sub_proto_module.yang_node_order_num = yang_node_order_num
        sub_proto_module.yang_node_name = container_yang_module.arg

        if container_yang_module.keyword == 'leaf-list':
            sub_proto_module.type = self.get_type_name_for_leaf(container_yang_module)
            if sub_proto_module.type == 'enum':
                sub_define_module = DefinionObj(proto_file_name)
                sub_define_module.keyword = 'enum'
                sub_define_module.parent = proto_module
                sub_define_module.name = sub_message_type
                sub_define_module.proto_xpath = '%s/%s' % (proto_module.proto_xpath, sub_message_type)
                sub_define_module.yang_node_order_num = container_yang_module.pos.line
                proto_module.definions[sub_message_type] = self.translate_enum_to_proto(
                    sub_define_module, sub_message_type, proto_file_name, container_yang_module)
                sub_proto_module.type = proto_module.definions[sub_message_type]
        else:
            sub_define_module = DefinionObj(proto_file_name)
            sub_define_module.keyword = 'message'
            sub_define_module.parent = proto_module
            sub_define_module.name = sub_message_type
            sub_define_module.proto_xpath = '%s/%s' % (proto_module.proto_xpath, sub_message_type)
            sub_define_module.yang_node_order_num = yang_node_order_num

            proto_module.definions[sub_message_type] = self._parse_statement(
                proto_file_name, sub_define_module, container_yang_module)
            sub_proto_module.type = proto_module.definions[sub_message_type]
        proto_module.statement.append(sub_proto_module)
        return proto_module

    def arrange_simple_type(self, leaf_yang_module, proto_file_name, proto_module):
        is_leaf_node = False
        if leaf_yang_module.keyword == 'leaf':
            is_leaf_node = True
        sub_message_name = self.trans_node_name(leaf_yang_module.arg, proto_module.proto_xpath, is_leaf_node)
        sub_proto_module = MessageObj(proto_file_name)
        sub_proto_module.scope = 'simple'
        sub_proto_module.name = sub_message_name
        sub_proto_module.parent = proto_module
        sub_proto_module.proto_xpath = '%s/%s' % (proto_module.proto_xpath, sub_message_name)
        sub_message_type = self.get_out_message_definion_name(leaf_yang_module.arg, proto_module.proto_xpath)
        yang_node_order_num = self.find_yang_node_order_num(leaf_yang_module)
        sub_proto_module.yang_node_order_num = yang_node_order_num
        sub_proto_module.yang_node_name = leaf_yang_module.arg
        if is_leaf_node:
            sub_proto_module = self.trans_yangleaf_to_proto(sub_proto_module, leaf_yang_module)
            if sub_proto_module.type == 'enum':
                sub_define_module = DefinionObj(proto_file_name)
                sub_define_module.keyword = 'enum'
                sub_define_module.parent = proto_module
                sub_define_module.name = sub_message_type
                sub_define_module.proto_xpath = '%s/%s' % (proto_module.proto_xpath, sub_message_type)
                sub_define_module.yang_node_order_num = leaf_yang_module.pos.line
                proto_module.definions[sub_message_type] = self.translate_enum_to_proto(
                    sub_define_module, sub_message_type, proto_file_name, leaf_yang_module)
                sub_proto_module.type = proto_module.definions[sub_message_type]
        elif leaf_yang_module.keyword in yang_keyword_not_support:
            return proto_module
        else:
            logging.info('%s is unreachable' % leaf_yang_module.keyword)
            sub_proto_module.type = sub_message_type
        proto_module.statement.append(sub_proto_module)
        return proto_module

    def arrange_oneof_type(self, container_yang_module, proto_file_name, proto_module):
        if container_yang_module.keyword in oneof_type:
            sub_message_name = self.trans_node_name(container_yang_module.arg, proto_module.proto_xpath)
            sub_message_type = self.get_out_message_definion_name(container_yang_module.arg, proto_module.proto_xpath)
            logging.info('generating oneof message node:%s', container_yang_module.arg)
            sub_proto_module = MessageObj(proto_file_name)
            sub_proto_module.scope = ''
            sub_proto_module.name = sub_message_name
            sub_proto_module.parent = proto_module
            sub_proto_module.proto_xpath = '%s' % proto_module.proto_xpath
            yang_node_order_num = self.find_yang_node_order_num(container_yang_module)
            sub_proto_module.yang_node_order_num = yang_node_order_num
            sub_proto_module.yang_node_name = container_yang_module.arg

            logging.info('arrange container type:yang_node:%s, define_node:%s, order num:%d'
                         % (container_yang_module.arg, sub_proto_module.name, sub_proto_module.yang_node_order_num))

            sub_define_module = DefinionObj(proto_file_name)
            sub_define_module.keyword = 'oneof'
            sub_define_module.parent = proto_module
            sub_define_module.name = sub_message_type
            sub_define_module.proto_xpath = '%s' % proto_module.proto_xpath
            sub_define_module.yang_node_order_num = yang_node_order_num

            logging.info('arrange container type:yang_node:%s, define_node:%s, order num:%d'
                         % (container_yang_module.arg, sub_define_module.name, sub_define_module.yang_node_order_num))

            proto_module.definions[sub_message_type] = self._parse_statement(
                proto_file_name, sub_define_module, container_yang_module)
            sub_proto_module.type = proto_module.definions[sub_message_type]
            proto_module.statement.append(sub_proto_module)
        return proto_module

    @staticmethod
    def calculate_proto_name(father_proto_xpath, proto_name, is_leaf_node=False):
        node_name_list = father_proto_xpath.upper().split('/')
        # leaf node can not be ext keep word, other structure can not be
        # ext keep words and origin keep words
        if proto_name.upper() in node_name_list:
            return '%s%s' %(proto_name, str(len(node_name_list)))
        if proto_name.replace('_', '').upper() in ext_keep_word:
            return '%s%s' % (proto_name, str(len(node_name_list)))
        if not is_leaf_node and proto_name.replace('_', '').upper() in origin_keep_word:
            return '%s%s' % (proto_name, str(len(node_name_list)))
        return proto_name

    def _parse_statement(self, proto_file_name, proto_node, yang_node):
        if proto_node.parent is None:
            proto_node.is_top = True
        split_node = self.split_yang_node(yang_node)
        for container_yang_node in split_node['message']:
            proto_node = self.arrange_message_type(container_yang_node, proto_file_name, proto_node)

        for container_yang_node in split_node['oneof']:
            proto_node = self.arrange_oneof_type(container_yang_node, proto_file_name, proto_node)

        for container_yang_node in split_node['repeated']:
            proto_node = self.arrange_repeated_type(container_yang_node, proto_file_name, proto_node)

        for leaf_yang_node in split_node['simple']:
            proto_node = self.arrange_simple_type(leaf_yang_node, proto_file_name, proto_node)

        return proto_node

    def translate_heads(self, yang_name, module, proto_file):
        for annotation in annotations:
            annotation_obj = AnnotationObj(yang_name, annotation, proto_file.pos.line)
            proto_file.i_annotations.append(annotation_obj)
        annotation_module = '// module %s\n' % module.arg
        annotation_obj = AnnotationObj(yang_name, annotation_module, proto_file.pos.line)
        proto_file.i_annotations.append(annotation_obj)
        revision_modules = module.search('revision')
        for revision_module in revision_modules:
            annotation_module = '// revision %s\n' % revision_module.arg
            annotation_obj = AnnotationObj(yang_name, annotation_module, proto_file.pos.line)
            proto_file.i_annotations.append(annotation_obj)
        namespace_modules = module.search('namespace')
        for revision_module in namespace_modules:
            annotation_module = '// namespace %s\n' % revision_module.arg
            annotation_obj = AnnotationObj(yang_name, annotation_module, proto_file.pos.line)
            proto_file.i_annotations.append(annotation_obj)

        proto_file.i_package = module.arg.replace('-', '_')
        return

    def set_annotations(self, annotation_obj, str, pos):
        annotation_obj.str = str
        annotation_obj.ref = pos



    def generate_proto_custom(self, yang_module):
        if not yang_module.i_children:
            return
        if not yang_module.keyword == 'module':
            return
        proto_file_name = os.path.join(self.out_proto_directory, ''.join([yang_module.arg, '.proto']))

        logging.info('generateing proto file:%s', proto_file_name)
        proto_file = ProtoFile(proto_file_name, proto_version)
        self.translate_heads(proto_file_name, yang_module, proto_file)
        self.translate_services()
        self.translate_messages(proto_file_name, yang_module, proto_file)
        return proto_file

    def get_yang_file_name(self, yang_directory, module_name):
        yang_name = ''.join([module_name, '.yang'])
        if yang_name in os.listdir(yang_directory):
            return os.path.join(yang_directory, yang_name)
        return ''

    def get_proto_file(self, proto_file_name, ptx):
        text = ''
        try:
            with open(proto_file_name, "r", encoding="utf-8") as fd:
                text = fd.read()
        except Exception:
            sys.stderr.write("can not open the file %s" % proto_file_name)
            logging.error("can not open the file %s" % proto_file_name)
        module = ptx.add_modules(proto_file_name, text)
        for error in ptx.errors:
            logging.info('%s in %s' % (str(str(error[2])), str(error[0])))
        if ptx.errors:
            logging.info("proto file %s parsing break." % proto_file_name)
            return None
        return module

    def get_module_for_out(self, src_proto_file_name, ptx, module_name, yang_proto_file):
        if os.path.isfile(src_proto_file_name):
            src_proto_module = self.get_proto_file(src_proto_file_name, ptx)
            if src_proto_module:
                for definion in src_proto_module.definions.values():
                    del_same_node_in_proto(definion)
            ret_val = merge_proto_module(yang_proto_file, src_proto_module)
            if ret_val == 0:
                m = proto_writer.Proto_writer(src_proto_module)
                return m
        m = proto_writer.Proto_writer(yang_proto_file)
        return m



    def translate_process(self, yang_module, ptx):
        proto_file = self.generate_proto_custom(yang_module)
        if not proto_file:
            return
        m = self.get_module_for_out(proto_file.ref, ptx, yang_module.arg, proto_file)
        if m:
            m.output_proto_file()

    def copy_old_proto(self):
        if not self.src_proto_directory or not os.path.isdir(self.src_proto_directory):
            logging.info('proto directory is null')
            return

        for proto_file in os.listdir(self.src_proto_directory):
            if not proto_file.endswith('.proto'):
                continue
            old_proto = os.path.join(self.src_proto_directory, proto_file)
            if proto_file not in os.listdir(self.out_proto_directory):
                shutil.copy(old_proto, self.out_proto_directory)
        return
    def file_is_exist(self,file,dir=None):
        if dir is None:
            return os.path.exists(file)
        if not os.path.isdir(dir):
            if os.path.samefile(file,dir):
                return True
            else:
                return False
        for sub_dir in os.listdir(dir):
            if os.path.isfile(sub_dir):
                if os.path.samefile(file,sub_dir):
                    return True
            else:
                if self.file_is_exist(file,os.path.join(dir,sub_dir)):
                    return True
        return False
    def start_translate(self):
        if not os.path.exists(self.log_directory):
            os.makedirs(self.log_directory)

        log_file_name = os.path.join(self.log_directory, ''.join([rep_revs_log_file, '.log']))
        if os.path.exists(log_file_name):
            os.remove(log_file_name)

        logging.basicConfig(filename=log_file_name,
                            format='%(asctime)s,%(levelname)s,%(filename)s,%(lineno)d:%(message)s',
                            level=logging.DEBUG,

                            )
        logging.info(
            "==============================START YANG BASE LINE PROCESS=========================================")
        logging.info("option is: \n yang_path: %s\n"
                     " log_file_name: %s\n"
                     " out_proto_directory: %s\n"
                     " src_proto_directory: %s\n"
                     " yang_dependencies: %s\n",
                     self.yang_path, log_file_name,
                     self.out_proto_directory, self.src_proto_directory,self.yang_dependencies,
                     )
        sys.stderr.write("start translating yang to proto \n")
        ptx = hpproto.Context()

        import platform
        if platform.system() == 'Windows':
            split_tag = ';'
        else:
            split_tag = ':'
        if os.path.isdir(self.yang_path):
            dir_yang_str = ''.join([str(self.yang_path), split_tag, str(self.yang_dependencies)])
            ctx = yang_util.init_ctx(dir_yang_str)
        else:
            ctx = yang_util.init_ctx(self.yang_dependencies)
        self.copy_old_proto()
        #parse yang path
        yang_util.parse_yang_modules(self.yang_path,ctx)
        if os.path.exists(self.yang_dependencies):
            yang_util.parse_yang_modules(self.yang_dependencies, ctx)
        if yang_util.yang_module_has_error(ctx):
            yang_util.print_yang_errors(ctx)
            return
        for module_name,rev in ctx.modules:
            yang_module = ctx.get_module(module_name,rev)
            src_yang_file = yang_module.pos.ref
            if self.file_is_exist(src_yang_file,self.yang_path):
                self.translate_process(yang_module, ptx)

        # if self.module_name:
        #     if yang_util.yang_module_has_error(ctx,self.module_name):
        #         yang_util.print_yang_errors(ctx,self.module_name)
        #         return
        #     self.translate_process(self.module_name, ptx, ctx)
        # else:
        #     yang_modules = yang_util.get_out_main_module_yang_from_path(self.yang_directory)
        #     for yang_module in yang_modules:
        #         if yang_util.yang_module_has_error(ctx, yang_module):
        #             yang_util.print_yang_errors(ctx, yang_module)
        #             return
        #         self.translate_process(yang_module, ptx, ctx)
        #     self.copy_old_proto()
        return ptx

def parse_error_and_exit(filename):
    ret_val = 0
    try:
        with open(filename) as handle:
            error_context = handle.read()
        findre = re.compile(r",ERROR," , re.DOTALL)
        error_list = findre.findall(error_context)
        if len(error_list) == 0:
            ret_val = 0
        else:
            ret_val = 1
            print('YANG_TO_PROTO_ERROR-----')
    except Exception:
        return 1
    finally:
        return ret_val


if __name__ == "__main__":
    start = time.time()
    options = env_parse()
    p = Yang_proto_converter(options)
    ptx = p.start_translate()
    sys.stderr.write("end process !\n")
    end = time.time()
    sys.stderr.write("process time:" +str(end - start))
    #log_file_name = os.path.join(os.path.abspath(options.log_directory), ''.join([rep_revs_log_file, '.log']))
    sys.exit(0)



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
# Copyright: (c) Huawei Technologies Co., Ltd. 2021. All rights reserved.


import os
from . import proto_parse
from . import proto_proto

__version__ = '1.1.0'
__date__ = '2021-08-02'


class Context(object):
    def __init__(self):
        self.errors = []

    def add_modules(self, ref, text):
        proto_module = proto_parse.ProtoParser()
        module = proto_module.parse(self, ref, text)
        if module is None:
            return None
        return module


from setuptools import setup

setup(
    name='protoyang',
    version='0.0.2',
    packages=['protoyang', 'protoyang.hpproto'],
    url='https://github.com/HuaweiDatacomm/protoyang',
    license='Apache License, Version 2.0',
    author='feng chong',
    author_email='frank.fengchong@huawei.com',
    description='a tool generate .proto files according yang modules'
)



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify
# it under the terms of the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License, Version 2.0 for more details.
import os
import io
import logging
import re
import sys
import traceback

from pyang.context import Context
from pyang.repository import FileRepository
import pyang

ERR_TYPE_GRAMMA = "Gramma"
ERR_TYPE_LOST = "Lost"
ERR_INFO = {
    # yang file it self has grammatical errors, use pyang check get detailed error information
    ERR_TYPE_GRAMMA: "***parse***: [yang file {0} has grammatical errors] Error 1",
    # yang module declared in white list, but correspond yang file not exists
    ERR_TYPE_LOST: "***check***: [yang file {0} doesn't exists] Erro 1",
}
ERR_REPORT_STAT = {
    ERR_TYPE_GRAMMA: [],
    ERR_TYPE_LOST: []
}
start_error = r"#########yang_Error_start##########"
end_error = r"#######yang_Error_end##########"


def init_ctx(path='./yang'):
    repos = FileRepository(path)
    ctx = Context(repos)
    return ctx


def parse_yang_module(yang_file_path, ctx):
    r = re.compile(r"^(.*?)(\@(\d{4}-\d{2}-\d{2}))?\.(yang|yin)$")
    try:
        with open(yang_file_path, "r", encoding="UTF-8") as fd:
            text = fd.read()
            m = r.search(yang_file_path)
            ctx.yin_module_map = {}
            if m is not None:
                (name, _dummy, rev, format) = m.groups()
                name = os.path.basename(name)
                ctx.add_module(yang_file_path, text, format, name, rev)
            else:
                ctx.add_module(yang_file_path, text)
    except Exception:
        yang_error_write("can not open the file %s" % yang_file_path)

    logging.info('parse yang module %s success!', yang_file_path)
    return ctx


def parse_yang_modules(yang_directory, ctx):
    if os.path.isfile(yang_directory):
        parse_yang_module(yang_directory, ctx)
    else:
        for yang_file in sorted(os.listdir(yang_directory)):
            yang_file_path = os.path.join(yang_directory, yang_file)
            if os.path.isdir(yang_file_path):
                parse_yang_modules(yang_file_path, ctx)
            parse_yang_module(yang_file_path, ctx)
    ctx.validate()
    return ctx

def get_moudle_name_from_annotations(annotations):
    for annotation_obj in annotations:
        pos = annotation_obj.str.find('module')
        if pos != -1:
            module_name = annotation_obj.str[pos + len('module'):].strip(' ').strip('\n')
            return module_name


def yang_module_has_error(ctx,module_name=None):
    for p,t,a in ctx.errors:
        if module_name is None:
            if is_error(t):
                return True
        else:
            if (is_error(t) and check_error_if_need(p, module_name)):
                return True

    return False

def check_error_if_need(pos, module_name):
    yang_error_name = ''
    try:
        yang_error_name = pos.top.arg
    except:
        pass
    if yang_error_name == module_name:
        return True
    return False

def is_error(error_type):
    error_level = pyang.error.err_level(error_type)
    if pyang.error.is_error(error_level):
        return True
    else:
        return False


def print_yang_errors(ctx, module_name=None):
    for p, t, a in ctx.errors:
        error_str = None
        error_level = pyang.error.err_level(t)
        if pyang.error.is_error(error_level):
            error_level_str = "Error"
        else:
            error_level_str = "Warning"
        if module_name is not None:
            if is_error(t) and check_error_if_need(p, module_name):
                error_str = ''.join([p.label(), " ", error_level_str,': ',pyang.error.err_to_str(t, a)])
        else:
            error_str = ''.join([p.label(), " ", error_level_str,': ',pyang.error.err_to_str(t, a)])
        if error_str is not None:
            print(error_str)
            yang_error_write(error_str)
    return

def yang_error_write(error_string):
    exc_info = sys.exc_info()
    exc_value = exc_info[1]

    if exc_value and traceback.format_exc().strip() != 'None':
        logging.error(traceback.format_exc())
    logging.error("%s",error_string)


def log_serious_errors(err_type, **kwargs):
    if err_type not in ERR_INFO:
        logging.info("receive unsupported error type")
        return

    if kwargs["file_name"] in ERR_REPORT_STAT[err_type]:
        return
    ERR_REPORT_STAT[err_type].append(kwargs["file_name"])
    logging.error(ERR_INFO[err_type].format(kwargs["file_name"]))


if __name__ == '__main__':
    pass



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
# Copyright: (c) Huawei Technologies Co., Ltd. 2021. All rights reserved.


import collections
import re
import sys
from hpproto.proto_proto import AnnotationObj
from hpproto.proto_proto import MessageObj
from hpproto.proto_proto import DefinionObj
from hpproto.proto_proto import RpcObj
from hpproto.proto_proto import ProtoFile
from . import syntax
import logging
import os
sys.path.append(os.path.abspath(os.path.dirname(__file__)+'/'+'..'))
from pyang import error

Proto_Base_Type = ['bytes','bool','string','int32','int64','uint32','uint64',
            'double','float','sint32','sint64','fixed32','sfixed32','fixed64','sfixed64', 'Decimal64']
Complex_Type =['enum','message','service','oneof']
Describe_Attr = ['repeated','optional','required','singular']
YANG_NODE_NAME_IN_PROTO_IDENTITY = 'json_name'


class ProtoTokenizer(ProtoFile):
    def __init__(self, text, ref, errors):
        super(ProtoTokenizer, self).__init__(ref, 'proto2')
        self.lines = collections.deque(text.splitlines(True))
        self.buf = ''
        self.offset = 0
        """Position on line.  Used to remove leading whitespace from strings."""
        self.errors = errors
        self.max_line = len(text.split('\n'))

    def readline(self):
        if len(self.lines) == 0:
            return
        self.buf = self.lines.popleft()
        self.pos.line += 1
        self.offset = 0

    def set_buf(self, pos):
        self.buf = self.buf[pos:]
        self.offset += pos

    def skip(self):
        buflen_before = len(self.buf)
        while True:
            self.buf = self.buf.lstrip()
            if self.buf:
                self.offset += (buflen_before - len(self.buf))
                break
            else:
                if len(self.lines) == 0:
                    return
                self.readline()
                buflen_before = len(self.buf)
        if self.buf[0] == '/':
            if self.buf[1] == '/':
                annotation = AnnotationObj(self.ref, self.buf, self.pos.line)
                self.i_annotations.append(annotation)
                if len(self.lines) == 0:
                    return
                self.readline()
                return self.skip()
            elif self.buf[1] == '*':
                anotation = ''
                line = self.pos.line
                anotation += self.buf
                i = self.buf.find('*/')
                while i == -1:
                    self.readline()
                    if len(self.lines) == 0:
                        return
                    anotation = ''.join([anotation, self.buf])
                    i = self.buf.find('*/')
                self.set_buf(i + 2)
                if self.pos.line < self.max_line:
                    annotation = AnnotationObj(self.ref, anotation, line)
                    self.i_annotations.append(annotation)
            return self.skip()

    def _parse_statement_type(self, args, definions):
        if args in Proto_Base_Type:
            type = args
        elif args in definions.keys():
            type = definions[args]
        else:
            error.err_add(self.errors, self.pos,
                          'SYNTAX_ERROR', 'illegal stmt: %s' % self.buf)
            raise error.Abort

        return type

    def parse_json_name(self, proto_str):
        yang_node_name = ''

        # 匹配old proto中[*]，将[*]当成json_name
        pattern_module = re.compile(r'(\[[\s\S]*\])')
        m = pattern_module.findall(proto_str.strip())
        if m:
            yang_node_name = m[0]
        return yang_node_name

    def get_proto_num_and_name(self, parent):
        proto_nums = []
        proto_names = []
        for stat in parent.statement:
            proto_nums.append(int(stat.num))
            proto_names.append(stat.name)
        return proto_nums, proto_names

    def validate_proto_num(self, proto_nums, node):
        max_num = -1
        if proto_nums:
            max_num = max(proto_nums)

        try:
            node_num = int(node.num)
            if node_num <= max_num:
                logging.warning('proto node:%s which num:%d less than before proto num in %s:%d'
                                % (node.name, node_num, self.pos.ref, self.pos.line))
        except:
            error.err_add(self.errors, self.pos,
                          'SYNTAX_ERROR', 'illegal proto num: %s' % node.num)
            raise error.Abort

    def validate_proto_name(self, proto_names, name):
        if name in proto_names:
            error.err_add(self.errors, self.pos,
                          'SYNTAX_ERROR', 'duplicate proto name: %s' % name)
            raise error.Abort

    def match_statement_content(self, proto_str):
        pattern_module = re.compile(r'(.*)\s*=\s*(\w*)\s*\[')
        match_content = pattern_module.findall(proto_str.strip())
        if match_content:
            return match_content
        else:
            pattern_module = re.compile(r'(.*)\s*=\s*(\w*)\s*;')
            return pattern_module.findall(proto_str.strip())

    def parse_augment(self, stmt):
        augment_file_name = os.path.basename(self.pos.ref).split('.')[0]
        if augment_file_name.endswith('_aug'):
            stmt.augmented = 1

    def parse_statement_content(self, parent):
        self.skip()
        if len(self.lines) == 0:
            return None
        if parent is None:
            parent = self
        split_strs = self.buf.split('//')
        proto_str = split_strs[0]
        proto_annotations = []
        if len(split_strs) > 1:
            proto_annotations = split_strs[1:]
        m = self.match_statement_content(proto_str)
        if m:
            arg = m[0][0].split()
            num = m[0][1]
            if not re.match(r'\d', num):
                error.err_add(self.errors, self.pos,
                              'SYNTAX_ERROR', 'illegal stmt: %s' % self.buf)
                raise error.Abort
            stmt = MessageObj(self.ref)
            stmt.num = num
            # enum
            if len(arg) == 1:
                stmt.scope = 'simple'
                stmt.name = arg[0]
            # string name
            elif len(arg) == 2:
                stmt.scope = 'simple'
                stmt.name = arg[1]
                if parent.keyword == 'oneof':
                    stmt.type = self._parse_statement_type(arg[0], parent.parent.definions)
                else:
                    stmt.type = self._parse_statement_type(arg[0], parent.definions)
            # repeated *** ***
            elif len(arg) == 3:
                if arg[0] not in Describe_Attr:
                    error.err_add(self.errors, self.pos,
                                  'SYNTAX_ERROR', 'illegal stmt: %s' % self.buf)
                    raise error.Abort
                stmt.scope = arg[0]
                stmt.type = self._parse_statement_type(arg[1], parent.definions)
                stmt.name = arg[2]
            else:
                error.err_add(self.errors, self.pos,
                              'SYNTAX_ERROR', 'illegal stmt: %s' % self.buf)
                raise error.Abort

            stmt.yang_node_name = self.parse_json_name(proto_str)
            stmt.proto_annotations = proto_annotations
            if parent:
                stmt.proto_xpath = '%s/%s' % (parent.proto_xpath, stmt.name)
            else:
                stmt.proto_xpath = '/%s' % stmt.name
            self.set_buf(len(self.buf.split(';')[0]))

            proto_nums, proto_names = self.get_proto_num_and_name(parent)
            self.validate_proto_num(proto_nums, stmt)
            self.validate_proto_name(proto_names, stmt.name)
            self.parse_augment(stmt)

            return stmt

        return None

    def get_keyword(self):
        r = syntax.re_keyword.match(self.buf)
        if r:
            self.set_buf(r.end())
            if not (self.buf[0] in (' ', ';', '{') or
                    self.buf[0:1] == '//' or self.buf[0:1] == '/*'):
                error.err_add(self.errors, self.pos,
                              'SYNTAX_ERROR', 'expected separator appear, but got: %s"..."' % self.buf[:6])
                raise error.Abort
            if r.group(2):
                return (r.group(2), r.group(3))
            else:
                return r.group(3)
        else:
            error.err_add(self.errors, self.pos,
                          'SYNTAX_ERROR', 'illegal keyword appear: ' + self.buf)
            raise error.Abort

    def peek(self):
        self.skip()
        return self.buf[0]

    def get_string(self, need_quote=False):
        """ret: string"""
        self.skip()

        if self.buf[0] in (';', '{', '}'):
            error.err_add(self.errors, self.pos,
                          'EXPECTED_ARGUMENT', self.buf[0])
            raise error.Abort
        while self.buf[0] in ('=', ' '):
            self.set_buf(1)
        if self.buf[0] in ('"', "'"):
            find_str = self.buf[0]
            buf_strs = []
            cur_pos = self.offset
            index = 1
            while True:
                buflen = len(self.buf)
                start_pos = index
                while index < buflen:
                    if self.buf[index] == find_str:
                        buf_strs.append(self.buf[start_pos:index])
                        self.set_buf(index + 1)
                        self.skip()
                        return ''.join(buf_strs)
                    index = index + 1
                buf_strs.append(self.buf[start_pos:index])
                self.readline()
                index = 0
                if find_str == '"':
                    buflen = len(self.buf)
                    while (index < buflen and self.buf[index] == '' and
                                   index <= cur_pos):
                        index = index + 1
                    if index == buflen:
                        index = 0
        elif need_quote:
            error.err_add(self.errors, self.pos, 'EXPECTED_QUOTED_STRING', ())
            raise error.Abort
        else:
            buflen = len(self.buf)
            index = 0
            while index < buflen:
                if (self.buf[index] in (';', '{', '}') or
                        self.buf[index:index + 2] == '//' or self.buf[index:index + 2] == '/*' or
                        self.buf[index:index + 2] == '*/'):
                    res = self.buf[:index]
                    self.set_buf(index)
                    return res
                index = index + 1

class ProtoParser(object):
    def __init__(self):
        pass

    def parse(self, ctx, ref, text):
        self.ctx = ctx
        try:
            self.tokenizer = ProtoTokenizer(text, ref, ctx.errors)
            self._parse_head()
            while (self.tokenizer.pos.line < self.tokenizer.max_line-1):
                self._parse_statement(None, ref)
        except error.Abort:
            # 解析不含message的proto文件时，清除此时报错，返回已解析的内容
            if self.tokenizer.buf == '' and len(self.tokenizer.i_messages) == 0:
                ctx.errors.clear()
                return self.tokenizer
            else:
                error.err_add(self.ctx.errors, self.tokenizer.pos, 'ABO_ERROR', ())
                return None
        except error.Eof as e:
            error.err_add(self.ctx.errors, self.tokenizer.pos, 'EOF_ERROR', ())
            return None
        return self.tokenizer

    def get_public_import(self, arg):
        imports = arg.split(' ')
        public_import = ''

        # proto头含有public的import：import public "aaa.proto";
        # 参考https://developers.google.com/protocol-buffers/docs/proto3
        if len(imports) == 2 and imports[0] == 'public':
            proto_file_name = imports[1].strip('"')
            public_import = '%spublic %s' % (public_import, proto_file_name)
        return public_import

    def validate_import(self, import_proto):
        proto_strs = import_proto.split('.')
        if len(proto_strs) == 2 and proto_strs[1] == 'proto':
            return True
        else:
            return False

    def _parse_head(self):
        self.tokenizer.skip()
        if len(self.tokenizer.lines) == 0:
            return None
        temp_buf = ''
        while True:
            temp_buf = self.tokenizer.buf
            keywd = self.tokenizer.get_keyword()
            tok = self.tokenizer.peek()
            if tok == '{' or tok == ';':
                arg = None
            else:
                arg = self.tokenizer.get_string().rstrip()
            if keywd in Complex_Type:
                self.tokenizer.buf = temp_buf
                return
            if keywd == 'syntax':
                self.tokenizer.i_syntax = arg
                if tok == '=':
                    self.tokenizer.skip()
                    self.tokenizer.set_buf(1)
                else:
                    error.err_add(self.ctx.errors, self.tokenizer.pos, 'INCOMPLETE_STATEMENT',
                                  (keywd, tok))
                    raise error.Abort
            # 例：import "***";
            elif keywd == 'import':
                import_proto_str = self.get_public_import(arg)
                if import_proto_str == '':
                    import_proto_str = arg

                if not self.validate_import(import_proto_str):
                    error.err_add(self.ctx.errors, self.tokenizer.pos, 'INVALID_IMPORT',
                                  (keywd, import_proto_str))
                    return error.Abort

                self.tokenizer.i_imports.append(import_proto_str)
                self.tokenizer.skip()
                self.tokenizer.set_buf(1)
            elif keywd == 'package':
                self.tokenizer.i_package = arg
                self.tokenizer.skip()
                self.tokenizer.set_buf(1)
            # rpc *** (***) returns (***);
            self.tokenizer.skip()

    def _parse_statement(self, parent, ref):
        is_definion_result = self.tokenizer.parse_statement_content(parent)
        if len(self.tokenizer.lines) == 0:
            self.tokenizer.pos.line += 1
            return None
        if is_definion_result:
            if self.tokenizer.peek() == ';':
                self.tokenizer.skip()
                self.tokenizer.set_buf(1)
            if parent is not None:
                is_definion_result.parent = parent
                parent.statement.append(is_definion_result)
                if parent.keyword == 'oneof' and parent.parent:
                    parent.parent.statement.append(is_definion_result)
            else:
                is_definion_result.is_top = True
                self.tokenizer.i_messages.append(is_definion_result)
        else:
            keywd = self.tokenizer.get_keyword()
            tok = self.tokenizer.peek()
            if tok == '{' or tok == ';':
                arg = None
            else:
                arg = self.tokenizer.get_string().rstrip()            
            if keywd == 'rpc':
                self.tokenizer.skip()
                self.tokenizer.set_buf(1)
                parent.statement.append(self.translate_rpc(parent, keywd, arg, ref))
            else:
                # 例：enum *** {*};
                stmt = self.translate_definion_type(ref, parent, keywd, arg)
                tok = self.tokenizer.peek()
                if tok == '{':
                    self.tokenizer.skip()
                    self.tokenizer.set_buf(1)  # skip the '{'
                    while self.tokenizer.peek() != '}':
                        if self.tokenizer.peek() == ';':
                            self.tokenizer.skip()
                            self.tokenizer.set_buf(1)
                        stmt = self._parse_statement(stmt, ref)
                    if parent is None:
                        if stmt.keyword == 'service':
                            self.tokenizer.i_services.append(stmt)
                        else:
                            self.tokenizer.definions[arg] = stmt
                    else:
                        parent.definions[arg] = stmt
                    self.tokenizer.skip()
                    self.tokenizer.set_buf(1)  # skip the '}'
                elif tok == ';':
                    self.tokenizer.skip()
                    self.tokenizer.set_buf(1)  # skip the ';'
                else:
                    error.err_add(self.ctx.errors, self.tokenizer.pos, 'INCOMPLETE_STATEMENT',
                                  (keywd, tok))
                    raise error.Abort
        return parent

    def translate_message_stmt(self, ref, parent, keywd, arg):
        # 例：string *** = *;
        if keywd in Proto_Base_Type:
            stmt = self.translate_base_type(ref, parent, keywd, arg)

        # 例：repeated *** *** = *;
        elif keywd in Describe_Attr:
            stmt= self.translate_describe_type(ref, parent, keywd, arg)
        # 例：*** *** = *;
        else:
            stmt = self.translate_custom_type(ref, parent, keywd, arg)

        return stmt

    def translate_definion_type(self, ref, parent, keywd, arg):
        stmt = DefinionObj(ref)
        stmt.parent = parent
        stmt.keyword = keywd
        stmt.name = arg
        stmt.pos.line = self.tokenizer.pos.line
        if parent is None:
            stmt.is_top = True
            stmt.proto_xpath = '/%s' % arg
        else:
            stmt.proto_xpath = '%s/%s' % (parent.proto_xpath, arg)
        return stmt

    def translate_describe_type(self, ref, parent, keywd, arg):
        stmt = MessageObj(ref)
        stmt.parent = parent
        stmt.scope = keywd
        stmt.num = arg.split('=')[1].lstrip()
        arg_tmp = arg.split('=')[0].rstrip()
        if len(arg_tmp.split(' ')) != 1:
            stmt.type = arg_tmp.split(' ')[0]
            stmt.name = arg_tmp.split(' ')[1]
        else:
            stmt.name = arg_tmp
        if parent is None:
            stmt.is_top = True
        stmt.pos.line = self.tokenizer.pos.line
        return stmt

    def translate_custom_type(self, ref, parent, keywd, arg):
        stmt = MessageObj(ref)
        stmt.parent = parent
        stmt.scope = 'simple'
        # 例：*** *** = *;
        if '=' in arg:
            try:
                stmt.type = parent.definions[keywd]
            except:
                stmt.type = keywd
                error.err_add(self.ctx.errors, self.tokenizer.pos, 'TYPE NOT FOUND',
                              (keywd ))
                raise error.Abort
            stmt.num = arg.split('=')[1].lstrip()
            stmt.name = arg.split('=')[0].rstrip()
        else:
            #例： *** = *;
            stmt.num = arg
            stmt.name = keywd
        if parent is None:
            stmt.is_top = True
        stmt.pos.line = self.tokenizer.pos.line
        return stmt

    def translate_base_type(self, ref, parent, keywd, arg):
        stmt = MessageObj(ref)
        stmt.parent = parent
        stmt.type = keywd
        stmt.scope = 'simple'
        stmt.num = arg.split('=')[1].lstrip()
        stmt.name = arg.split('=')[0].rstrip()
        if parent is None:
            stmt.is_top = True
        stmt.pos.line = self.tokenizer.pos.line
        return stmt

    def translate_rpc(self, parent, keywd, arg, ref):
        stmt = RpcObj(ref)
        stmt.parent = parent
        stmt.keyword = keywd
        stmt.name = arg.split(' ')[0]
        stmt.Request = arg[len(arg.split('(')[0])+1:len(arg.split(')')[0])]
        Response = arg.split('returns')[1]
        stmt.Response = Response[len(Response.split('(')[0])+1:len(Response.split(')')[0])]
        if parent is None:
            stmt.is_top = True
        stmt.pos.line = self.tokenizer.pos.line
        return stmt


#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
#Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.

#This program is free software; you can redistribute it and/or modify
#it under the terms of the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.

#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#Apache License, Version 2.0 for more details.
import logging
import os
from hpproto.proto_proto import DefinionObj


def get_proto_parents(proto_msg, xpaths):
    if proto_msg.parent is not None:
        xpaths.insert(0, proto_msg.parent.name)
        get_proto_parents(proto_msg.parent, xpaths)


def get_proto_xpaths(proto_msg):
    xpaths = []
    get_proto_parents(proto_msg, xpaths)
    xpath_str = ''
    for xpath in xpaths:
        xpath_str = '%s/%s' % (xpath_str, xpath)

    xpath_str = '%s/%s' % (xpath_str, proto_msg.name)
    return xpath_str


# 1.old proto没有enum_val值，取num值作为enum_val
def get_old_proto_enum_value(proto_msg):
    try:
        if proto_msg.parent.keyword == 'enum'and proto_msg.enum_val == -1:
            proto_msg.enum_val = int(proto_msg.num)
    except:
        logging.info('get enum value from old proto error, proto name:%s, xpath:%s'
                      % (proto_msg.name, get_proto_xpaths(proto_msg)))


def get_msg_type(msg):
    if isinstance(msg.type, DefinionObj):
        return msg.type.keyword
    else:
        return msg.type


def is_proto_type_equal_to_yang_type(proto_msg, yang_msg):
    if isinstance(proto_msg.type, DefinionObj) and isinstance(yang_msg.type, DefinionObj):
        if proto_msg.type.keyword == yang_msg.type.keyword: #and proto_msg.type.name == yang_msg.type.name:
            return True
        else:
            return False
    elif isinstance(proto_msg.type, DefinionObj):
        return False
    elif isinstance(yang_msg.type, DefinionObj):
        return False
    else:
        if proto_msg.type != yang_msg.type:
            logging.info('old proto node type not equal to yang node type, node name:%s, '
                            'old proto type:%s, yang proto type:%s' % (proto_msg.name, proto_msg.type, yang_msg.type))
            if proto_msg.type is None:
                proto_msg.type = yang_msg.type
        return proto_msg.type == yang_msg.type


def compare_enum_name_ignore_case(proto_msg, yang_msg):
    proto_msg_parent = proto_msg.parent
    yang_msg_parent = yang_msg.parent
    if proto_msg_parent.name == yang_msg_parent.name and proto_msg_parent.keyword == 'enum' and \
       yang_msg_parent.keyword == 'enum':
        proto_enum = proto_msg.name.lower()
        yang_enum = yang_msg.name.lower()
        if proto_enum == yang_enum:
            if proto_msg.name != yang_msg.name:
                logging.warning('old proto enum name not equal to yang enum, proto enum:%s, yang enum:%s'
                                % (proto_msg.name, yang_msg.name))
            return True
    return False


def find_proto_msg(proto_definion, yang_msg):
    for proto_msg in proto_definion.statement:
        if proto_msg.name == yang_msg.name:
            proto_msg.scope = yang_msg.scope
            if proto_msg.yang_node_name == '':
                proto_msg.yang_node_name = yang_msg.yang_node_name
            if is_proto_type_equal_to_yang_type(proto_msg, yang_msg):
                get_old_proto_enum_value(proto_msg)
            else:
                logging.info('old proto node type not equal to yang node type, node name:%s, xpath:%s, '
                              'old proto type:%s, yang node type:%s, out yang type.'
                              % (proto_msg.name, get_proto_xpaths(proto_msg), get_msg_type(proto_msg), get_msg_type(yang_msg)))

                # raise Exception('old proto type not equal to yang type')

                if isinstance(yang_msg.type, DefinionObj) and yang_msg.type.keyword == 'oneof':
                    for sub_proto_definion in proto_definion.definions.keys():
                        if proto_msg.type == proto_definion.definions[sub_proto_definion]:
                            del proto_definion.definions[sub_proto_definion]
                            break
                proto_msg.type = yang_msg.type
            return proto_msg
        elif compare_enum_name_ignore_case(proto_msg, yang_msg):  # old proto enum名为小写，比较时不区分大小写，认为是相同的
            return proto_msg

    return None


def find_proto_definion(proto_definions, yang_definion, recursive=False):
    if len(proto_definions) > 0:
        for obj in proto_definions.values():
            if obj.name == yang_definion.name:
                return obj
            else:
                if recursive:
                    find_proto_definion(obj.definions, yang_definion, True)


def find_yang_definion_existed_in_proto_definion(proto_definion, yang_definion):
    return find_proto_definion(proto_definion.definions, yang_definion, True)


# if any value in yang enum type, use yang value num
def judge_use_old_enum(yang_definion):
    for yang_msg in yang_definion.statement:
        if yang_msg.enum_val != -1:
            return False
    return True


def _compare_statement(yang_definion, proto_definion):

    if not judge_use_old_enum(yang_definion):
        # ignore old proto enum structure
        proto_definion.statement = yang_definion.statement
        return

    for yang_msg in yang_definion.statement:
        proto_msg = find_proto_msg(proto_definion, yang_msg)
        if proto_msg is None:
            logging.info("statement nonexistent %s" % yang_msg.name)

            # check if proto file has old definions of msg obj with statement none.
            if isinstance(yang_msg.type, DefinionObj):
                def_obj = find_yang_definion_existed_in_proto_definion(proto_definion, yang_msg.type)
                if def_obj is not None:
                    yang_msg.type = def_obj
                else:
                    logging.info("def obj nonexistent %s" % yang_msg.type.name)

            proto_definion.statement.append(yang_msg)


def _compare_definions(yang_definions, proto_definions, proto_parent_definion=None):
    if len(yang_definions) > 0:
        for yang_definion in yang_definions.values():
            proto_definion = find_proto_definion(proto_definions, yang_definion)
            if proto_definion is not None:
                _compare_definions(yang_definion.definions, proto_definion.definions, proto_definion)
                _compare_statement(yang_definion, proto_definion)
            else:
                logging.info('ignore definion %s' % yang_definion.name)
                proto_definions[yang_definion.name] = yang_definion
                if proto_parent_definion is not None:
                    proto_parent_definion.definions = proto_definions


def _replace_annotations(yang_proto, proto):
    proto.i_annotations = yang_proto.i_annotations


def del_same_node_in_proto(module):
    del_definions = []
    del_statements = []
    if isinstance(module, DefinionObj) and not judge_xpath_has_same_node(module.proto_xpath):
        del_definions.append(module.name)
        return del_definions

    for definion in list(module.definions.values()):
        if not judge_xpath_has_same_node(definion.proto_xpath):
            del_definions.append(definion.name)
            logging.info('proto path have same node name! %s'% definion.proto_xpath)
        if definion.definions:
            for child_definion in list(definion.definions.values()):
                del_definions.extend(del_same_node_in_proto(child_definion))
    for del_definion in del_definions:
        if del_definion not in module.definions.keys():
            continue
        module.definions.pop(del_definion)

    for statement in list(module.statement):
        if not judge_xpath_has_same_node(statement.proto_xpath):
            del_statements.append(statement)
            logging.info('proto path have same node name! %s' % statement.proto_xpath)
        if isinstance(statement.type, DefinionObj):
            if del_same_node_in_proto(statement.type):
                del_statements.append(statement)
    for del_stat in del_statements:
        if del_stat not in module.statement:
            continue
        module.statement.remove(del_stat)
    return []


def judge_xpath_has_same_node(xpath):
    node = xpath.split('/')
    if len(node) == len(set(node)):
        return True
    return False


def merge_proto_module(yang_module, proto_module):
    try:
        if yang_module is not None and proto_module is not None:
            yang_definions = yang_module.definions
            proto_definions = proto_module.definions
            _compare_definions(yang_definions, proto_definions)
            _replace_annotations(yang_module, proto_module)
        else:
            if yang_module is None:
                logging.info('proto node translated from yang file is None')
            elif proto_module is None:
                logging.info('old proto is empty, out yang module')
                return 1
    except Exception:
        logging.info('can not merge old proto, old proto: %s, new proto: %s'
                      % (os.path.basename(proto_module.pos.ref), os.path.basename(yang_module.pos.ref)))
        return 1
    return 0



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
# Copyright: (c) Huawei Technologies Co., Ltd. 2021. All rights reserved.

import sys
import os
sys.path.append(os.path.abspath(os.path.dirname(__file__)+'/'+'..'))
from pyang import error

KEYWORD = ['message', 'enum', 'oneof', 'service','rpc']
SCOPE = ['repeated', 'optional', 'required', 'simple', '']


class DefinionObj(object):
    def __init__(self, ref):
        self.keyword = ''  # message enum oneof service
        self.pos = error.Position(ref)
        self.definions = {}
        self.statement = []
        self.is_print = False
        self.parent = None
        self.is_top = False
        self.name = ''
        self.augmented = 0  # 是否是augment的节点：1是，0否
        self.yang_node_order_num = 0  # 记录yang节点的序号，输出proto文件时按该序号排序
        self.proto_xpath = ''


class MessageObj(object):
    def __init__(self, ref):
        self.parent = None
        self.scope = ''  # repeated optional required simple null
        self.type = None
        self.name = ''
        self.num = ''
        self.pos = error.Position(ref)
        self.is_top = False
        self.augmented = 0  # 是否是augment的节点：1是，0否
        self.yang_node_order_num = 0  # 记录yang节点的序号，输出proto文件时按该序号排序
        self.yang_node_name = ''  # 记录proto节点对应yang文件节点名，输出proto时使用，如果oldproto该值为空，则替换；否则，不作改动
        self.enum_val = -1  # 记录proto节点对应yang文件中enum value值，有old proto时，取enum对应的proto num
        self.proto_annotations = []  # 记录old proto中节点对应注释信息
        self.proto_xpath = ''


class RpcObj(object):
    def __init__(self, ref):
        self.parent = None
        self.is_top = False
        self.scope = ''
        self.keyword = 'rpc'
        self.type = ''
        self.Request = ''
        self.Response = ''
        self.name = ''
        self.num = ''
        self.pos = error.Position(ref)
        self.proto_xpath = ''


class AnnotationObj(object):
    def __init__(self, ref, string, index):
        self.str = string
        self.pos = error.Position(ref)
        self.pos.line = index


class ProtoFile(object):
    def __init__(self,ref,version):
        self.i_syntax = version
        self.i_imports = []
        self.i_annotations = []
        self.i_package = ''
        self.i_services = []
        self.i_messages = []
        self.i_root_elements = []
        self.definions = {}
        self.ref = ref
        self.pos = error.Position(ref)





#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
# Copyright: (c) Huawei Technologies Co., Ltd. 2021. All rights reserved.
"""Description of PROTO syntax."""

import re

### Regular expressions - constraints on arguments

# keywords and identifiers
identifier = r"[_A-Za-z][._\-A-Za-z0-9]*"
keyword = '((' + identifier + '):)?(' + identifier + ')'
comment = '(/\*([^*]|[\r\n\s]|(\*+([^*/]|[\r\n\s])))*\*+/)|(//.*)|(/\*.*)'
num = r'0-9'

re_keyword = re.compile(keyword)
re_comment = re.compile(comment)